import os, json, shutil
from pathlib import Path
from lxml import etree
import markdown

ABSOLUTES = ('/', '//', 'http://', 'https://')
EXTERNALS = ('//', 'http://', 'https://')
LINK_ATTRIBUTES = ('href', 'src')

# parses an XML file but adds a prologue to handle non-breaking spaces
def parse_with_prologue(path):
	print(path)
	with open(path, "r", encoding="utf-8") as file:
		return etree.fromstring("<!DOCTYPE xml [<!ENTITY nbsp ' '>]>" + file.read())

# writes HTML files, adding the doctype declaration
def write_with_prologue(root, path):
	with open(path, "w", encoding="utf-8") as file:
		file.write("<!DOCTYPE html>\n\n" + etree.tostring(root, encoding="unicode", method='html'))

# removes everything within a given directory
def empty(path):
	for file_path in path.iterdir():
		if file_path.is_dir():
			empty(file_path)
			os.rmdir(file_path)
		else:
			os.remove(file_path)

# copies everything inside the source directory to the config directory, and
# processes any .3sg files it comes across
def compile(source, destination, config):
	for source_path in source.iterdir():
		destination_path = destination / source_path.name
		if source_path.is_dir():
			os.mkdir(destination_path)
			compile(source_path, destination_path, config)
		else:
			if source_path.suffix == ".3sg":
				compile_file(source_path, destination_path.with_suffix(''), config)
			else:
				shutil.copyfile(source_path, destination_path)

# processes the .3sg file at source, and sends the output to destination
def compile_file(source, destination, config):
	# get the source and check it is a <body> file. skip if not
	source_root = parse_with_prologue(source)
	if source_root.tag != 'body':
		return

	# process special tags (<include>, <md>)
	process_py(source_root, source, config['default_parameters'], config)
	process_includes(source_root, source, config['default_parameters'], config)
	process_md(source_root, source, config)

	# process broken links
	process_links(source_root, source, config)

	# create the destination tree
	html_root = etree.Element('html')
	head_element = etree.SubElement(html_root, 'head')
	html_root.append(source_root)

	# add title, favicon, stylesheets, scripts, and meta elements
	if 'title' in source_root.attrib:
		title_element = etree.SubElement(head_element, 'title')
		title_element.text = source_root.attrib['title']
		source_root.attrib.pop('title')
	if 'favicon' in source_root.attrib:
		icon_element = etree.SubElement(head_element, 'link', rel='icon', href=source_root.attrib['favicon'])
		source_root.attrib.pop('favicon')
	if 'stylesheet' in source_root.attrib:
		for link in source_root.attrib['stylesheet'].split():
			stylesheet_element = etree.SubElement(head_element, 'link', rel='stylesheet', href=link)
		source_root.attrib.pop('stylesheet')
	if 'script' in source_root.attrib:
		for link in source_root.attrib['script'].split():
			stylesheet_element = etree.SubElement(head_element, 'script', src=link)
		source_root.attrib.pop('script')
	charset_element = etree.SubElement(head_element, 'meta', charset="UTF-8")
	viewport_element = etree.SubElement(head_element, 'meta', name="viewport", content="width=device-width,initial-scale=1.0")

	# write to destination
	if config['indexify'] and os.path.basename(source) != "index.3sg":
		# under indexify, we must add a ../ to all links
		for attribute in LINK_ATTRIBUTES:
			for link in html_root.findall(f".//*[@{attribute}]"):
				if not link.attrib[attribute].startswith(ABSOLUTES):
					link.attrib[attribute] = str(Path("..") / link.attrib[attribute]).replace("\\", "/")

		os.mkdir(destination)
		etree.indent(html_root)
		write_with_prologue(html_root, os.path.join(destination, "index.html"))
	else:
		etree.indent(html_root)
		write_with_prologue(html_root, destination.with_suffix(".html"))

# processes a .3sg file at include_source which is to placed within the file at
# source. does not do link processing, but will repair links and do markdown
def include_file(source, include_source, parameters, config):
	# get the include source
	include_root = parse_with_prologue(include_source)

	# process special tags (<include>, <md>). we can ignore broken links until
	# later, at the very end once everything is repaired and such
	process_py(include_root, include_source, parameters, config)
	process_includes(include_root, include_source, parameters, config)
	process_md(include_root, include_source, config)

	# repair links: essentially we can't just concatenate the link from source
	# to include_source with that from include_source to link_source, as we
	# would get superfluous '..'s. Instead, we must find the absolute path to
	# link_source, and find the shortest relative path from source. Exception
	# if the link has the 'stay_rel' attribute.
	for attribute in LINK_ATTRIBUTES:
		for link in include_root.findall(f".//*[@{attribute}]"):
			if not link.attrib[attribute].startswith(ABSOLUTES) and not 'stay_rel' in link.attrib:
				link_source = include_source.parent / link.attrib[attribute]
				link.attrib[attribute] = str(link_source.resolve().relative_to(source.parent.resolve(), walk_up=True)).replace("\\", "/")

	# return the root element
	return include_root

# finds all <py> tags and executes them with the current set of local variables
def process_py(source_root, source, parameters, config):
	for py_tag in source_root.iter('py'):
		# process the python
		local_parameters = parameters.copy()
		exec(py_tag.text, {}, local_parameters)
		python_root = etree.fromstring("<py>" + local_parameters['pyout'] + "</py>")
		
		# replace the <py> tag
		py_tag.tag = "div"
		if 'class' in py_tag.attrib:
			py_tag.attrib['class'] += " python"
		else:
			py_tag.attrib['class'] = "python"
		py_tag.text = python_root.text
		for child in python_root:
			py_tag.append(child)

# finds all <include> tags and replaces them with the processed contents of the
# target .3sg file
def process_includes(source_root, source, parameters, config):
	for include_tag in source_root.iter('include'):
		# get the path to the file to be included
		if include_tag.attrib['src'].startswith(ABSOLUTES):
			include_source = (config['source'] / Path(include_tag.attrib['src']).relative_to(config['parent'])).with_suffix(".3sg")
		else:
			include_source = (source.parent / include_tag.attrib['src']).with_suffix(".3sg")

		# create parameters
		local_parameters = parameters.copy()
		for attribute in include_tag.attrib:
			if attribute.startswith("para-"):
				local_parameters[attribute[5:]] = include_tag.attrib[attribute]
				include_tag.attrib.pop(attribute)

		# process the included .3sg file
		include_root = include_file(source, include_source, local_parameters, config)

		# replace the <include> tag
		include_tag.tag = include_root.tag
		include_tag.attrib.pop('src')
		include_tag.attrib.update(include_root.attrib)
		for child in include_root:
			include_tag.append(child)

# finds all <md> tags and replaces them with a <div> with the appropriate content
def process_md(source_root, source, config):
	for md_tag in source_root.iter('md'):
		# process the markdown
		markdown_root = etree.fromstring("<md>" + markdown.markdown(md_tag.text, extensions=['tables', 'fenced_code']) + "</md>")

		# replace the <md> tag
		md_tag.tag = "div"
		if 'class' in md_tag.attrib:
			md_tag.attrib['class'] += " markdown"
		else:
			md_tag.attrib['class'] = "markdown"
		md_tag.text = ""
		for child in markdown_root:
			md_tag.append(child)

# iterates through all <a> tags and does things if the don't link to another page
def process_links(source_root, source, config):
	for link_tag in source_root.iter('a'):
		# a link is allowed to pass if it meets one of these criteria:
		# 1. it doesn't point anywhere
		if 'href' not in link_tag.attrib:
			continue

		# 2. it points outside the website
		if link_tag.attrib['href'].startswith(EXTERNALS):
			continue

		# 3. it points to somewhere outside the scope of the simpleSSG compiler
		# at this point we can also determine where the link points to in the source directory
		if link_tag.attrib['href'].startswith('/'):
			if not Path(link_tag.attrib['href']).is_relative_to(config['parent']):
				continue

			link_destination = config['source'] / Path(link_tag.attrib['href']).relative_to(config['parent'])
		else:
			link_destination = source.parent / link_tag.attrib['href']

			if not link_destination.resolve().is_relative_to(config['source'].resolve()):
				continue

		# 4. it points to a file
		if link_destination.is_file() and link_destination.suffix != ".3sg":
			continue

		# 5. it points to a directory which has an index.html or index.htm file.
		if link_destination.is_dir() and ((link_destination / "index.html").is_file() or (link_destination / "index.htm").is_file()):
			continue

		# 6. it points to a directory which has an index.3sg <body> file
		if link_destination.is_dir() and (link_destination / "index.3sg").is_file():
			source_root = parse_with_prologue(link_destination / "index.3sg")
			if source_root.tag == 'body':
				continue

		# 7. (under indexify) it points to a non-existant directory for which there is a matching .3sg <body> file
		if config['indexify'] and link_destination.suffix == "" and link_destination.with_suffix(".3sg").is_file():
			source_root = parse_with_prologue(link_destination.with_suffix(".3sg"))
			if source_root.tag == 'body':
				continue

		# 8. (without indexify) it points to a non-existant file for which there is a matching .3sg <body> file
		if not config['indexify'] and link_destination.suffix == ".html" and link_destination.with_suffix(".3sg").is_file():
			source_root = parse_with_prologue(link_destination.with_suffix(".3sg"))
			if source_root.tag == 'body':
				continue

		# if none of those conditions are met, it means the link is broken
		# thus, add a class and/or remove the href attributes
		if config['broken_class']:
			if 'class' in link_tag.attrib:
				link_tag.attrib['class'] += " " + config['broken_class']
			else:
				link_tag.attrib['class'] = config['broken_class']
		if config['remove_href']:
			del link_tag.attrib['href']

if __name__ == '__main__':
	with open("config.cfg") as config_file:
		config = json.loads(config_file.read())
	config['source'] = Path(config['source'])
	config['destination'] = Path(config['destination'])
	config['parent'] = Path(config['parent'])

	if config['empty_destination']:
		empty(config['destination'])
	compile(config['source'], config['destination'], config)